import sys

from loguru import logger


# Remove all pre-configured loggers. We just want the ones we're about to add
logger.remove(None)

# Add a simple logger with no fluff.
logger.add(sys.stderr, format="<yellow>{time:HH:mm:ss}</yellow> {message}", colorize=True)

# All other files will import from here.
log = logger.opt(colors=True)



from blackbox.handlers.databases._base import BlackboxDatabase
from blackbox.utils import run_command
from blackbox.utils.logger import log


class MongoDB(BlackboxDatabase):
    """
    A Database handler that will do a mongodump for MongoDB, backing up all documents.

    This will use mongodump with --gzip and --archive, and must be restored using the same
    arguments, e.g. mongorestore --gzip --archive=/path/to/file.archive.
    """

    required_fields = ("connection_string",)
    backup_extension = ".archive"

    def backup(self, backup_path) -> None:
        """Dump all the data to a file and then return the filepath."""
        # Run the backup, and store the outcome in this object.
        self.success, self.output = run_command(
            f"mongodump "
            f"--uri={self.config['connection_string']} "
            "--gzip "
            "--forceTableScan "
            f"--archive={backup_path}"
        )
        log.debug(self.output)



import requests

from blackbox.handlers.notifiers._base import BlackboxNotifier


class Discord(BlackboxNotifier):
    """A notifier for sending webhooks to Discord."""

    required_fields = ("webhook",)

    def _parse_report(self) -> dict:
        """Turn the report into something the notify function can use."""
        # Combine and truncate total output to < 2000 characters, fields don't support more.
        output = self.report.output[:2000]

        # Was this a success?
        success = self.report.success
        color = 1024049 if success else 13377568

        # Make a list of database fields
        fields = []
        for database in self.report.databases:
            field = {
                "name": f"**{database.database_id}**",
                "inline": True,
                "value": ""
            }

            for provider in database.storages:
                emoji = ":white_check_mark:" if provider.success else ":x:"
                field['value'] += f"{emoji}  {provider.storage_id}\n"

            # If all backup fails, no storage statuses will be added to
            # database['storage']. Discord doesn't allow empty field
            # values, so we have to add the :x: emoji to the field to
            # prevent the webhook endpoint from raising an error.
            if not field['value']:
                field['value'] = ":x:"

            # Strip any trailing newlines and append
            field['value'] = field['value'].strip()
            fields.append(field)

        # Add the output, but only if the overall report is a failure.
        if not success:
            fields.append({
                "name": "Output",
                "value": output,
            })

        # Return the payload
        return {
            "content": None,
            "embeds": [
                {
                    "title": "Backup",
                    "color": color,
                    "fields": fields
                }
            ],
            "username": "blackbox",
            "avatar_url": "https://raw.githubusercontent.com/lemonsaurus/blackbox/main/img/blackbox_avatar.png"  # NOQA: E501
        }

    def notify(self):
        """Send a webhook to Discord with a blackbox report."""
        requests.post(self.config["webhook"], json=self._parse_report())



import typing as t


class ImproperlyConfigured(Exception):
    """A class for raising configuration errors in our handlers."""


class MissingFields(ImproperlyConfigured):
    """Exception for missing configuration fields in handlers."""

    def __init__(self, handler: str, name: str, id_: t.Optional[str], fields: list[str]):
        id_info = f" with id {id_}" if id_ is not None else ""
        field_info = ', '.join(fields)
        self.message = f"Missing configuration key(s) for {handler} {name}{id_info}: {field_info}"
        super().__init__(self.message)


class InvalidHandler(ImproperlyConfigured):
    """
    Exception for configured handlers that do not exist.

    Example: Defining a database "Postman" or a storage "Droopbox"
    """

    def __init__(self, handler: str):
        self.message = f"Invalid handler: {handler}"
        super().__init__(self.message)


class UnknownId(ImproperlyConfigured):
    """
    Exception for configured ids that do not exist.

    Example: Specifying a storage with id `main_bucket` when that id does not exist
             in the `storages` section
    """

    def __init__(self, database: str, database_id: str, id_: str):
        self.message = (
            f"Unknown id specified for database {database} with id {database_id}: {id_}"
        )
        super().__init__(self.message)



import os
import pathlib

import yaml
from jinja2 import Environment
from jinja2 import StrictUndefined
from jinja2.exceptions import TemplateSyntaxError
from jinja2.exceptions import UndefinedError

from blackbox.exceptions import ImproperlyConfigured


def get_yaml_config(config_path: pathlib.Path) -> dict:
    """Load the yaml file in the root folder, and return it as a dict."""
    template = Environment(undefined=StrictUndefined)

    try:
        with open(config_path, encoding="UTF-8", mode="r") as f:
            # Inject the local environment variables into the rendering context.
            # This bit of magic allows us to resolve any {{ VARIABLE }} to whatever
            # the value of the equivalent environment variable is.
            parsed_config = template.from_string(f.read()).render(**os.environ)
            return yaml.safe_load(parsed_config)

    except TemplateSyntaxError as e:
        raise ImproperlyConfigured(
            f"There is an error in the config file: {e}"
        ) from e

    except UndefinedError as e:
        raise ImproperlyConfigured(
            f"Missing environment variable: {e}"
        ) from e

    except FileNotFoundError as e:
        raise ImproperlyConfigured(
            f"Your blackbox.yaml file was not found at {config_path}!\n"
            "See the readme under 'Configuration' for more information."
        ) from e



from pathlib import Path
from zipfile import ZIP_DEFLATED
from zipfile import ZipFile

from blackbox.exceptions import ImproperlyConfigured
from blackbox.handlers.databases._base import BlackboxDatabase


class LocalStorage(BlackboxDatabase):
    """A Database handler that will zip a local folder."""

    required_fields = ("path",)

    def __init__(self, **kwargs) -> None:
        # Gzip deflates only accept compression level ranging from 0 to 9
        # if the argument is given, we check it is in the allowed range otherwise
        # RuntimeError will be raised by ZipFile
        if compression_level := kwargs.get("compression_level"):
            if compression_level < 0 or compression_level > 9:
                raise ImproperlyConfigured(
                    f"Invalid compression level. "
                    f"Must be an integer between 0 and 9, got {compression_level}."
                )

        super().__init__(**kwargs)

    def backup(self, backup_path: Path) -> None:
        path = self.config["path"]
        compression_level = self.config.get("compression_level", 5)

        # Store evey file in the archive
        # We use deflate (Gzip) for compression and the level has already been validated in __init__
        with ZipFile(backup_path, "w", ZIP_DEFLATED, compresslevel=compression_level) as zipfile:
            for subpath in Path(path).rglob("*"):
                if subpath.is_file():
                    zipfile.write(subpath)

        # The compression was successful
        self.success = True



from abc import abstractmethod
from pathlib import Path

from blackbox.handlers._base import BlackboxHandler


class BlackboxDatabase(BlackboxHandler):
    """An abstract database handler."""

    handler_type = "database"
    backup_extension = ""

    def __init__(self, **kwargs):
        """Set up database handler."""
        super().__init__(**kwargs)

        self.success = False  # Was the backup successful?
        self.output = ""      # What did the backup output?

    @abstractmethod
    def backup(self, backup_path: Path):
        """
        Back up a database to the provided backup Path.

        All subclasses must implement this method.
        """
        raise NotImplementedError

    @property
    def output(self):
        """Return sanitized output only."""
        return self.__output

    @output.setter
    def output(self, sensitive_output: str):
        """Set sanitized output."""
        self.__output = self.sanitize_output(sensitive_output)

    def get_id_for_retention(self) -> str:
        """Used for deleting only this kind of old backups."""
        return self.config.get("id")



import os
import re
from datetime import datetime
from pathlib import Path

from dropbox import Dropbox as DropboxClient
from dropbox.exceptions import ApiError
from dropbox.exceptions import AuthError
from dropbox.exceptions import HttpError
from dropbox.files import CommitInfo
from dropbox.files import FileMetadata
from dropbox.files import UploadSessionCursor
from dropbox.files import WriteMode

from blackbox.config import Blackbox
from blackbox.handlers.storage._base import BlackboxStorage
from blackbox.utils.logger import log


class Dropbox(BlackboxStorage):
    """Storage handler that uploads backups to Dropbox."""

    required_fields = ("access_token",)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.upload_base = self.config.get("upload_directory") or "/"
        self.client = DropboxClient(self.config["access_token"])
        self.valid = self._validate_token()

    def _validate_token(self):
        """Check if dropbox token is valid."""
        try:
            return self.client.check_user("test").result == "test"
        except AuthError:
            return False

    def sync(self, file_path: Path) -> None:
        """Sync a file to Dropbox."""
        # Check if Dropbox token is valid.
        if self.valid is False:
            error = "Dropbox token is invalid!"
            self.success = False
            self.output = error
            log.error(error)
            return None

        # This is size what can be uploaded as one chunk.
        # When file is bigger than that, this will be uploaded
        # in multiple parts.
        chunk_size = 4 * 1024 * 1024

        temp_file, recompressed = self.compress(file_path)
        upload_path = f"{self.upload_base}{file_path.name}{'.gz' if recompressed else ''}"

        try:
            with temp_file as f:
                file_size = os.stat(f.name).st_size
                log.debug(file_size)
                if file_size <= chunk_size:
                    self.client.files_upload(
                        f.read(), upload_path, WriteMode.overwrite
                    )
                else:
                    session_start = self.client.files_upload_session_start(
                        f.read(chunk_size)
                    )
                    cursor = UploadSessionCursor(
                        session_start.session_id,
                        offset=f.tell()
                    )
                    # Commit contains path in Dropbox and write mode about file
                    commit = CommitInfo(upload_path, WriteMode.overwrite)

                    while f.tell() < file_size:
                        if (file_size - f.tell()) <= chunk_size:
                            self.client.files_upload_session_finish(
                                f.read(chunk_size),
                                cursor,
                                commit
                            )
                        else:
                            self.client.files_upload_session_append(
                                f.read(chunk_size),
                                cursor.session_id,
                                cursor.offset
                            )
                            cursor.offset = f.tell()
            self.success = True
        except (ApiError, HttpError) as e:
            log.error(e)
            self.success = False
            self.output = str(e)

    def rotate(self, database_id: str) -> None:
        """
        Rotate the files in the Dropbox directory.

        All files in base directory of backups will be deleted when they
        are older than `retention_days`, and because of this,
        it's better to have backups in isolated folder.
        """
        # Check if Dropbox token is valid.
        if self.valid is False:
            log.error("Dropbox token is invalid - Can't delete old backups!")
            return None
        # Let's rotate only this type of database
        db_type_regex = rf"{database_id}_blackbox_\d{{2}}_\d{{2}}_\d{{4}}.+"

        # Receive first batch of files.
        files_result = self.client.files_list_folder(
            self.upload_base if self.upload_base != "/" else ""
        )
        entries = [entry for entry in files_result.entries if
                   self._is_backup_file(entry, db_type_regex)]

        # If there is more files, receive all of them.
        while files_result.has_more:
            cursor = files_result.cursor
            files_result = self.client.files_list_folder_continue(cursor)
            entries += [entry for entry in files_result.entries if
                        self._is_backup_file(entry, db_type_regex)]

        retention_days = 7
        if Blackbox.retention_days:
            retention_days = Blackbox.retention_days

        # Find all old files and delete them.
        for item in entries:
            last_modified = item.server_modified
            now = datetime.now(tz=last_modified.tzinfo)
            delta = now - last_modified
            if delta.days >= retention_days:
                self.client.files_delete(item.path_lower)

    @staticmethod
    def _is_backup_file(entry, db_type_regex) -> bool:
        """Check if file is actually this kind of database backup."""
        return isinstance(entry, FileMetadata) and re.match(db_type_regex, entry.name)



import datetime
import os
import re
from pathlib import Path

import boto3
from botocore.exceptions import BotoCoreError
from botocore.exceptions import ClientError

from blackbox.config import Blackbox
from blackbox.exceptions import ImproperlyConfigured
from blackbox.handlers.storage._base import BlackboxStorage
from blackbox.utils.logger import log


class S3(BlackboxStorage):
    """A storage handler for S3-compatible APIs."""

    required_fields = ("bucket", "endpoint")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.bucket = self.config["bucket"]
        self.endpoint = self.config["endpoint"]

        # If the optional parameters for credentials have been provided, we use these.
        key_id = self.config.get("aws_access_key_id")
        secret_key = self.config.get("aws_secret_access_key")
        configuration = dict()

        # If config was provided for both of these, we should use it!
        if key_id and secret_key:
            configuration = {
                "aws_access_key_id": key_id,
                "aws_secret_access_key": secret_key,
            }

        # If config was provided for only one of them, that's too weird of a state for us to accept,
        # so we'll raise an exception. (That weird ^ operator is an XOR).
        elif bool(key_id) ^ bool(secret_key):
            raise ImproperlyConfigured(
                "You must configure either both or none of the S3 credential params.")

        else:
            # If config hasn't been provided, we expect either environment variables or ~/.aws/
            # credentials and config files to exist. If none of that exists, we should raise
            # a convenient error.
            has_environment_variables = (
                os.environ.get("AWS_ACCESS_KEY_ID")
                and os.environ.get("AWS_SECRET_ACCESS_KEY")
            )
            has_aws_config_files = (
                (Path.home() / ".aws/config").exists()
                and (Path.home() / ".aws/credentials").exists()
            )
            if not has_aws_config_files and not has_environment_variables:
                raise ImproperlyConfigured(
                    "Blackbox could not find any valid S3 credentials. "
                    "See the readme under Configuration for more information on how to do this."
                )

        # If we get to this point, the user has either environment variables or credentials files,
        # so Blackbox will make use of these.
        # See https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html
        self.session = boto3.Session(
            **configuration
        )
        self.client = self.session.client(
            's3',
            endpoint_url=f"https://{self.endpoint}",
        )

    def sync(self, file_path: Path) -> None:
        """Sync a file to an S3 bucket."""
        file_, recompressed = self.compress(file_path)

        try:
            self.client.upload_fileobj(
                file_,
                self.bucket,
                f"{file_path.name}{'.gz' if recompressed else ''}",
                ExtraArgs={"ContentEncoding": "gzip"}
            )
            self.success = True
        except (ClientError, BotoCoreError) as e:
            log.error(e)
            self.output = str(e)
            self.success = False

    def rotate(self, database_id: str) -> None:
        """
        Rotate the files in the S3 bucket.

        This deletes all files older than `config.BlackBox.retention_days` days old, as long as
        those files fit certain regular expressions. We don't want to delete
        files that are not related to backup or logging.
        """
        db_type_regex = rf"{database_id}_blackbox_\d{{2}}_\d{{2}}_\d{{4}}.+"
        retention_days = 7
        if Blackbox.retention_days:
            retention_days = Blackbox.retention_days

        # Get files object from remote S3 bucket.
        remote_objects = self.client.list_objects_v2(Bucket=self.bucket).get("Contents")

        # Filter their names with only this kind of database.
        relevant_backups = [item for item in remote_objects
                            if re.match(db_type_regex, item.get("Key"))]

        # Look through the items and figure out which ones are older than `retention_days`.
        # Catch all boto errors and log them to avoid return code 1.
        try:
            for item in relevant_backups:
                last_modified = item.get("LastModified")
                now_tz = datetime.datetime.now(tz=last_modified.tzinfo)
                delta = now_tz - item.get("LastModified")

                # Delete the deprecated items
                if delta.days >= retention_days:
                    self.client.delete_object(
                        Bucket=self.bucket,
                        Key=item.get("Key")
                    )
        except (ClientError, BotoCoreError) as e:
            log.error(e)



import telebot
from telebot.apihelper import ApiTelegramException

from blackbox.handlers.notifiers._base import BlackboxNotifier
from blackbox.utils.logger import log


STRING_LIMIT = 2000  # Limit output
CHECKMARK_EMOJI = "\U00002705"  # ✔
FAIL_EMOJI = "\U0000274C"       # ❌
WARNING_EMOJI = "\U000026A0"    # ⚠


class Telegram(BlackboxNotifier):
    """Telegram notifier for Blackbox."""

    required_fields = ("token", "chat_id",)

    def _parse_report(self) -> str:
        """Convert the report object to a Telegram-friendly string."""
        data_report = "Blackbox Backup Status:\n"

        for db in self.report.databases:
            data_report += f"{db.database_id}: \n"
            for storage in db.storages:
                if storage.success:
                    data_report += f"{CHECKMARK_EMOJI} {storage.storage_id}\n"
                else:
                    data_report += f"{FAIL_EMOJI} {storage.storage_id}\n"
            if db.success is False:
                data_report += f"{WARNING_EMOJI} {db.output[:STRING_LIMIT]}\n"
        return data_report

    def notify(self):
        """Convert report dict to string and send via Telegram."""
        bot = telebot.TeleBot(self.config["token"])
        try:
            bot.send_message(
                chat_id=self.config["chat_id"],
                text=self._parse_report(),
            )
        except ApiTelegramException:
            log.debug(
                "Telegram API key or user_id is wrong "
                "or you forgot to press /start in your bot."
            )



from blackbox.handlers.databases._base import BlackboxDatabase
from blackbox.utils import run_command
from blackbox.utils.logger import log


class Redis(BlackboxDatabase):
    """A Database handler that will run a redis-cli command for Redis backup."""

    required_fields = ("password", "host", )
    backup_extension = ".rdb"

    def backup(self, backup_path) -> None:
        """Dump all the data to a file and then return the filepath."""
        # Run the backup, and store the outcome.
        self.success, self.output = run_command(
            "redis-cli "
            f"-h {self.config.get('host')} "
            f"-p {str(self.config.get('port', '6379'))} "
            f"--rdb {backup_path}",
            REDISCLI_AUTH=self.config.get("password")
        )
        log.debug(self.output)



from abc import abstractmethod

from blackbox.handlers._base import BlackboxHandler
from blackbox.utils import reports


class BlackboxNotifier(BlackboxHandler):
    """An abstract interface for creating Blackbox Notifiers."""

    handler_type = "notifier"

    def __init__(self, **kwargs):
        """Set up notifier handler."""
        super().__init__(**kwargs)
        self.report = reports.Report()

    @abstractmethod
    def _parse_report(self) -> dict:
        """Turn the report into something the notify function can use."""
        raise NotImplementedError

    @abstractmethod
    def notify(self):
        """Send a notification to the configured notifier."""
        raise NotImplementedError

    def add_database(self, report: reports.DatabaseReport):
        """Add a database report to the current list of reports."""
        self.report.databases.append(report)



import requests

from blackbox.handlers.notifiers._base import BlackboxNotifier


class Slack(BlackboxNotifier):
    """A notifier for sending webhooks to Slack."""

    required_fields = ("webhook",)

    def _parse_report(self) -> dict:
        """Turn the report from main.py into something the notify function can use."""
        if self.config.get("use_block_kit"):
            return self._parse_report_modern()

        return self._parse_report_classic()

    def _parse_report_classic(self) -> dict:
        """Turn the report from main.py into Slack webhook payload with secondary attachment."""
        attachment = {
            "mrkdwn_in": ["fields"],
            "title": "Backup",
            "author_name": "blackbox",
            "author_icon": "https://raw.githubusercontent.com/lemonsaurus/blackbox/main/img/blackbox_avatar.png"  # NOQA: E501
        }

        # Combine and truncate total output to < 2000 characters, fields don't support more.
        output = self.report.output[:2000]

        # Was this a success?
        success = self.report.success
        attachment["color"] = "#0FA031" if success else "#CC2020"

        # Make a list of database fields
        fields = []
        for database in self.report.databases:
            field = {
                "title": f"{database.database_id}",
                "short": True,
                "value": ""
            }

            for provider in database.storages:
                emoji = ":white_check_mark:" if provider.success else ":x:"
                field['value'] += f"{emoji}  {provider.storage_id}\n"

            # Indicate that backup have failed if no storage providers was used.
            if not field['value']:
                field['value'] = ":x:"

            # Strip any trailing newlines and append
            field['value'] = field['value'].strip()
            fields.append(field)

        if not success:
            fields.append({
                "title": "Output",
                "value": output
            })

        attachment["fields"] = fields

        return {
            "attachments": [attachment]
        }

    def _parse_report_modern(self) -> dict:
        """Turn the report from main.py into Slack webhook payload with Block Kit."""
        blocks = [
            {
                "type": "header",
                "text": {
                    "type": "plain_text",
                    "text": "Backup"
                }
            }
        ]

        # Was this a success?
        success = self.report.success
        databases = list(self.report.databases)

        for i in range(0, len(databases), 2):
            dbs = [databases[i]]
            if len(databases) > i + 1:
                dbs.append(databases[i + 1])

            fields = []
            for db in dbs:
                field = {
                    "type": "mrkdwn",
                    "text": f"*{db.database_id}*"
                }

                for provider in db.storages:
                    emoji = ":white_check_mark:" if provider.success else ":x:"
                    field["text"] += f"\n{emoji} {provider.storage_id}"

                # Indicate that backup have failed if no storage providers was used.
                if field["text"] == f"*{db.database_id}*":
                    field["text"] += "\n:x:"

                # Strip any trailing newlines and append
                field['text'] = field['text'].strip()
                fields.append(field)

            section = {
                "type": "section",
                "fields": fields
            }
            blocks.append(section)

        if not success:
            # Combine and truncate total output to < 2000 characters, fields don't support more.
            output = self.report.output[:2000]
            blocks += [
                {
                    "type": "divider"
                },
                {
                    "type": "header",
                    "text": {
                        "type": "plain_text",
                        "text": "Output"
                    }
                },
                {
                    "type": "section",
                    "text": {
                        "type": "plain_text",
                        "text": output
                    }
                }
            ]

        blocks.append(
            {
                "type": "context",
                "elements": [
                    {
                        "type": "image",
                        "image_url": "https://raw.githubusercontent.com/lemonsaurus/blackbox/main/img/blackbox_avatar.png",  # NOQA: E501
                        "alt_text": "blackbox"
                    },
                    {
                        "type": "plain_text",
                        "text": "blackbox",
                        "emoji": True
                    }
                ]
            }
        )

        return {"blocks": blocks}

    def notify(self) -> None:
        """Send a webhook to Slack with a blackbox report."""
        requests.post(self.config["webhook"], json=self._parse_report())



import os
from pathlib import Path

from blackbox.utils.logger import log
from blackbox.utils.yaml import get_yaml_config


class YAMLGetter(type):
    """
    Metaclass for accessing yaml blackbox data.

    Supports getting configuration from up to two levels
    of nested configuration through `section` and `subsection`.
    - `section` specifies the YAML configuration section
      (or "key") in which the configuration lives.
    - `subsection` specifies the section within the
      section from which configuration should be loaded.

    If neither are these are set, it'll just try to get it from the root level.

    Adapted from python-discord/bot.

    Example Usage:
        # blackbox.yml
        bot:
            prefixes:
                direct_message: ''
                guild: '!'

        # blackbox.py
        class Prefixes(metaclass=YAMLGetter):
            section = "bot"
            subsection = "prefixes"

        # Usage in Python code
        from blackbox import Prefixes

        print(Prefixes.direct_message)
        print(Prefixes.guild)
    """

    section = None
    subsection = None
    _config = None

    @classmethod
    def parse_config(cls, config_path: Path = None):
        """Parse the config from the blackbox.yaml file."""
        # Get the config path from environment variables
        env_config_path = os.environ.get("BLACKBOX_CONFIG_PATH")

        # If the config path is given from elsewhere use that path
        if config_path:
            cls._config = get_yaml_config(config_path)

        # If there's an environment variable with a path, we'll use that.
        elif env_config_path:
            cls._config = get_yaml_config(Path(env_config_path))

        # Otherwise, we expect the config file to be in the root folder,
        # and to be called 'blackbox.yaml'
        else:
            root_folder = Path(__file__).parent.parent.absolute()
            cls._config = get_yaml_config(root_folder / "blackbox.yaml")

    def _get_annotation(cls, name):
        """Fetch the annotation configured in the subclass."""
        return cls.__annotations__.get(name)

    def __getattr__(cls, name):
        """
        Fetch the attribute in the `YAMLGetter.config` dictionary.

        This just attempts to do a dictionary lookup in `YAMLGetter.config`,
        and supports sections and subsections if we need nested blackbox data.
        """
        name = name.lower()

        # Here's a bit of magic - We don't set the cls._config until the first time
        # something tries to actually call this method. This kind of just-in-time approach
        # allows us to comfortably modify the `_config` at some point before we start using
        # any of the Handlers, without worrying about any race conditions.
        #
        # If you need this config to be set before or after the first call to __getattr__,
        # simply call YAMLGetter.parse_config().
        # You can pass in a configuration file path if you need to.
        if not cls._config:
            cls.parse_config()

        try:
            if cls.section is None:
                return cls._config[name]
            elif cls.subsection is None:
                return cls._config[cls.section][name]
            else:
                return cls._config[cls.section][cls.subsection][name]
        except KeyError:
            # If one of the handler lists isn't defined, return an empty list.
            log.warning(
                f"{name} is not defined in the blackbox.yaml file -- "
                "returning an falsy value."
            )
            if cls._get_annotation(name) == list:
                return []
            elif cls._get_annotation(name) == dict:
                return {}
            else:
                return None

    def __getitem__(cls, name):
        """Just defer to the __getattr__ implementation."""
        return cls.__getattr__(name)

    def __iter__(cls):
        """Return generator of key: value pairs of current constants class' blackbox values."""
        for name in cls.__annotations__:
            yield name, getattr(cls, name)


class Blackbox(metaclass=YAMLGetter):
    """The configuration for the blackbox application."""

    # Handlers
    databases: dict
    storage: dict
    notifiers: dict

    # Configuration
    retention_days: int
    notifier_frequency: int



from blackbox.handlers.databases._base import BlackboxDatabase
from blackbox.utils import run_command
from blackbox.utils.logger import log


class Postgres(BlackboxDatabase):
    """A Database handler that will do a pg_dumpall for Postgres, backing up all tables."""

    required_fields = ("username", "password", "host", )
    backup_extension = ".sql"

    def backup(self, backup_path) -> None:
        """Dump all the data to a file and then return the filepath."""
        # Run the backup, and store the outcome.
        self.success, self.output = run_command(
            f"pg_dumpall --file={backup_path}",
            PGUSER=self.config["username"],
            PGPASSWORD=self.config["password"],
            PGHOST=self.config["host"],
            PGPORT=str(self.config.get("port", "5432")),
        )
        log.debug(self.output)



import os
import subprocess
from typing import Tuple

from blackbox.utils.logger import log


def run_command(command: str, **environment) -> Tuple[bool, str]:
    """
    Execute the command, and log the result.

    Any additional keyword arguments passed into this call
    will be added as environment variables.

    Returns a tuple of (success, output), where success is a boolean value
    that is either True or False, and output is a string.
    """
    # Get the current environment variables.
    env = os.environ.copy()

    # Now add all the environment variables we passed in that are not None.
    extra_env = {key: value for key, value in environment.items() if value is not None}
    env.update(extra_env)

    # Run the command and capture the output
    try:
        result = subprocess.run(
            [command],
            shell=True,
            env=env,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )
        output = result.stdout.decode("utf-8").strip()
        success = True
    except subprocess.CalledProcessError as e:
        output = e.stdout.decode("utf-8").strip()
        success = False

    # Log and return output
    log.info(output)
    return success, output



import dataclasses
import typing as t
from collections import defaultdict
from itertools import chain

from blackbox import exceptions
from blackbox.handlers import BlackboxDatabase
from blackbox.handlers import BlackboxNotifier
from blackbox.handlers import BlackboxStorage
from blackbox.handlers._base import BlackboxHandler


Handler = t.TypeVar("Handler", bound=BlackboxHandler)
HandlerById = t.Mapping[str, set[Handler]]


@dataclasses.dataclass
class Workflow:
    """Dataclass that stores a database and the storage and notifiers specified for each."""

    database: BlackboxDatabase
    storage_providers: set[BlackboxStorage] = dataclasses.field(default_factory=set)
    notifiers: set[BlackboxNotifier] = dataclasses.field(default_factory=set)


# Get a mapping of handler names to their class objects ("postgres" to Postgres)
# Example:
#   HANDLER_MAPPING = { "postgres": Postgres, "discord": Discord }
components = (BlackboxDatabase, BlackboxNotifier, BlackboxStorage)

HANDLER_MAPPING: dict[str, t.Type[BlackboxHandler]] = {
    handler.__name__.lower(): handler
    for handler in chain.from_iterable(component.__subclasses__() for component in components)
}


def get_configured_handlers(config: dict) -> dict:
    """
    Instantiate handlers based on the given configuration.

    Returns a dictionary mapping handler ids to handler instances.
    Adds additional ids "all" and <handler_type> (eg. postgres, s3).
    """
    handler_dict = defaultdict(set)

    # Get handlers
    for handler_type, handler_config in config.items():
        try:
            Handler = HANDLER_MAPPING[handler_type]
        except KeyError:
            raise exceptions.InvalidHandler(handler_type)

        for handler_id, handler_fields in handler_config.items():
            handler_instance = Handler(**handler_fields, id=handler_id)

            # Set of all handlers
            handler_dict["all"].add(handler_instance)
            # Set of all handler type, ie. postgres/s3/discord
            handler_dict[handler_type].add(handler_instance)
            # Set of handlers with the specified id
            handler_dict[handler_id].add(handler_instance)

    return handler_dict


def get_handlers_by_id(
    id_: t.Union[str, list[str]],
    handlers: HandlerById[Handler]
) -> set[Handler]:
    """
    Given ids and a mapping of id to handlers, return handlers matching the ids.

    `id_` can be a string or a list of strings corresponding to ids.
    Raises TypeError if id_ is not a string or a list.
    Raises KeyError if id_ does not correspond to a handler in `handlers`.
    """
    if not isinstance(id_, (str, list)):
        raise TypeError

    # Turn string into a list with one string for simple iteration
    ids = [id_] if isinstance(id_, str) else id_

    match = set()
    for i in ids:
        match.update(handlers[i])
    return match


def get_workflows(
    database_handlers: HandlerById[BlackboxDatabase],
    storage_handlers: HandlerById[BlackboxStorage],
    notifier_handlers: HandlerById[BlackboxNotifier],
) -> list[Workflow]:
    """Return a list of workflows based on configured handlers."""
    workflows = []

    for database in database_handlers["all"]:
        workflow = Workflow(database)

        try:
            # Assign storage providers
            wanted_providers = database.config.get("storage_providers", "all")
            storage_providers = get_handlers_by_id(wanted_providers, storage_handlers)
            workflow.storage_providers.update(storage_providers)

            # Assign notifiers
            wanted_notifiers = database.config.get("notifiers", "all")
            notifiers = get_handlers_by_id(wanted_notifiers, notifier_handlers)
            workflow.notifiers.update(notifiers)
        except TypeError:
            raise exceptions.ImproperlyConfigured(
                "storage_providers and notifiers have to be a list of ids or an id"
            )
        except KeyError as e:
            raise exceptions.UnknownId(
                database.__class__.__name__, database.config["id"], e.args[0]
            )

        workflows.append(workflow)
    return workflows



from blackbox.handlers.databases._base import BlackboxDatabase
from blackbox.utils import run_command
from blackbox.utils.logger import log


class MariaDB(BlackboxDatabase):
    """A Database handler that will do a mysqldump for MariaDB, backing up all tables."""

    required_fields = ("username", "password", "host", )
    backup_extension = ".sql"

    def backup(self, backup_path) -> None:
        """Dump all the data to a file and then return the filepath."""
        user = self.config["username"]
        password = self.config["password"]
        host = self.config["host"]
        port = str(self.config.get("port", "3306"))

        # Run the backup, and store the outcome.
        self.success, self.output = run_command(
            f"mysqldump -h {host} -u {user} --password='{password}' "
            f"--port={port} --all-databases > {backup_path}"
        )
        log.debug(self.output)
        # Explicitly check if error message is occurred.
        # Somehow mysqldump is always successful.
        if "error" in self.output.lower():
            self.success = False
            log.debug("mysqldump has error(s) in log")



import gzip
import shutil
import tempfile
import typing
from abc import abstractmethod
from pathlib import Path

from blackbox.handlers._base import BlackboxHandler
from blackbox.utils.logger import log


# File suffixes considered as an archive
ARCHIVE_SUFFIXES = {".tar", ".zip"}


class BlackboxStorage(BlackboxHandler):
    """An abstract interface for creating Blackbox Storage Providers."""

    handler_type = "storage"

    def __init__(self, **kwargs):
        """Set up storage handler."""
        super().__init__(**kwargs)

        self.success = False  # Was the upload successful?
        self.output = ""     # What did the storage upload output?

    @staticmethod
    def compress(file_path: Path) -> tuple[typing.IO, bool]:
        """
        Compress the file using gzip into a tempfile.TemporaryFile.

        Returns a two elements tuple.
        The first one is a file-like object, which is removed when it is closed.
        The second one is True if the file has been recompressed, False otherwise.

        This should always be called before syncing the
        file to a storage provider.
        """
        # If the file is already considered as an archive, we don't recompress it
        if file_path.suffix in ARCHIVE_SUFFIXES:
            log.debug(f"File {file_path.name} is already compressed.")
            return open(file_path, "rb"), False

        temp_file = tempfile.NamedTemporaryFile(suffix=f"-{file_path.name}")

        log.debug(f"Compressing to temporary file: {temp_file.name}")
        with file_path.open(mode="rb") as f_in:
            with gzip.open(temp_file, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

        temp_file.seek(0)
        return temp_file, True

    @abstractmethod
    def sync(self, file_path: Path):
        """
        Sync a file to a storage provider.

        All subclasses must implement this method.
        """
        raise NotImplementedError

    @abstractmethod
    def rotate(self, database_id: str):
        """
        Rotate the files in the storage provider.

        This deletes all files older than `rotation_days` days old, as long as
        those files fit certain regular expressions. We don't want to delete
        files that are not related to backup or logging.

        All subclasses must implement this method.
        """
        raise NotImplementedError



import dataclasses

from blackbox.utils.mixins import SanitizeReportMixin


@dataclasses.dataclass
class StorageReport:
    """A report created by one of the Storage Handlers."""

    storage_id: str
    success: bool


@dataclasses.dataclass
class DatabaseReport(SanitizeReportMixin):
    """Keep database report."""

    database_id: str
    success: bool
    output: str
    storages: list[StorageReport] = dataclasses.field(default_factory=list)

    def report_storage(self, storage_id: str, success: bool, output: str):
        """Add a storage report to the current report."""
        # Add to database output
        self.output += self.sanitize_output(output)

        # If one storage handler failed, the database backup is a failure.
        if success is False:
            self.success = False

        # Add report to list of storages
        report = StorageReport(storage_id, success)
        self.storages.append(report)


@dataclasses.dataclass
class Report(SanitizeReportMixin):
    """Keep combined report."""

    databases: list[DatabaseReport] = dataclasses.field(default_factory=list)

    @property
    def success(self) -> bool:
        """Return whether or not the workflow is a success."""
        return all(report.success for report in self.databases)

    @property
    def output(self) -> str:
        """Return the combined outputs from all the database reports."""
        return self.sanitize_output("\n".join(report.output for report in self.databases))

    @property
    def is_empty(self) -> bool:
        """Return whether or not there are no databases in the report."""
        return self.databases == []



"""
Blackbox is a plug-and-play service which magically backs up all your databases.

The backups are stored on your favorite cloud storage providers, and Blackbox will notify
you on your chat platform of choice once the job is done.
"""
import datetime
import logging
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent

import click

from blackbox import exceptions
from blackbox.config import Blackbox as CONFIG
from blackbox.config import YAMLGetter
from blackbox.utils import workflows
from blackbox.utils.cooldown import is_on_cooldown
from blackbox.utils.logger import log
from blackbox.utils.reports import DatabaseReport


def run() -> bool:
    """
    Implement the main blackbox process.

    Returns whether or not the process is a success.
    """
    # Overall program success
    success = True

    # Parse configuration and instantiate handlers
    if not CONFIG.databases or not CONFIG.storage:
        raise exceptions.ImproperlyConfigured("You have to define least one database and storage")

    database_handlers = workflows.get_configured_handlers(CONFIG.databases)
    storage_handlers = workflows.get_configured_handlers(CONFIG.storage)
    notifier_handlers = workflows.get_configured_handlers(CONFIG.notifiers)

    all_workflows = workflows.get_workflows(database_handlers, storage_handlers, notifier_handlers)

    with TemporaryDirectory() as backup_dir:
        log.info(f"Backing up to folder: {backup_dir}")
        backup_dir = Path(backup_dir)
        date = datetime.date.today().strftime("%d_%m_%Y")
        backup_files = []

        for workflow in all_workflows:
            database = workflow.database

            # Do a backup, then return the path to the backup file
            backup_filename = f"{database.config['id']}_blackbox_{date}{database.backup_extension}"
            backup_path = backup_dir / backup_filename
            database.backup(backup_path)
            backup_files.append(backup_path)
            database_id = database.get_id_for_retention()
            database.teardown()

            # Add report to notifiers
            report = DatabaseReport(database.config["id"], database.success, database.output)
            for notifier in workflow.notifiers:
                notifier.add_database(report)

            # If backup failed, continue to next database. No need to sync.
            if not database.success:
                continue

            for storage in workflow.storage_providers:
                # Sync the provider, then rotate and cleanup
                storage.sync(backup_path)
                storage.rotate(database_id)
                storage.teardown()

                # Store the outcome to the database report
                report.report_storage(storage.config["id"], storage.success, storage.output)

            # Set overall program success to False if workflow is unsuccessful
            if report.success is False:
                success = False

        cooldown = CONFIG['cooldown']
        logging.debug(f"Cooldown setting is {cooldown}")
        if cooldown:
            is_on_cooldown_ = is_on_cooldown(cooldown)

        # Send a report for each notifier configured
        for notifier in notifier_handlers["all"]:
            # Don't send a notification if no database uses the notifier
            if notifier.report.is_empty:
                continue

            # If cooldown is not set or if report is failed: just notify.
            if not notifier.report.success:
                log.debug("Backup failed, sending notification.")
                notifier.notify()
            elif cooldown is None:
                log.debug('Cooldown config is None, sending notification.')
                notifier.notify()

            # But otherwise let's check do we have a right to notify
            else:
                if not is_on_cooldown_:
                    notifier.notify()

            notifier.teardown()
        return success


@click.command()
@click.option('--config', help="Path to blackbox.yaml file.")
@click.option('--init', is_flag=True, help="Generate blackbox.yaml file and exit.")
@click.option('--version', is_flag=True, help="Show version and exit.")
def cli(config, init, version):
    """
    BLACKBOX

    Backup database to external storage system
    """  # noqa
    if version:
        from blackbox import __version__
        print(f"Blackbox {__version__}")
        exit()

    if init:
        config_file = Path("blackbox.yaml")
        if not config_file.exists():
            config_file.write_text(dedent(
                """
                databases:
                  mongodb:
                    main_mongodb:
                      connection_string: mongodb://username:password@host:port
                  postgres:
                    main_postgres:
                      username: username
                      password: password
                      host: host
                      port: port
                  redis:
                    main_redis:
                      password: password
                      host: host
                      port: port

                storage:
                  s3:
                    main_s3:
                      bucket: bucket
                      endpoint: s3.endpoint.com
                      aws_access_key_id: dancinglemon
                      aws_secret_access_key: lemondance

                notifiers:
                  discord:
                    test_server:
                      webhook: https://web.hook/

                retention_days: 7
                """).lstrip())
            print("blackbox.yaml configuration created", flush=True)

        else:
            print("blackbox.yaml already exists", flush=True)

        exit()

    if config:
        YAMLGetter.parse_config(Path(config))

    success = run()
    # If this failed, we'll exit with a non-zero exit code, to indicate
    # a failure. Might be useful for Kubernetes jobs.
    if not success:
        exit(1)



from blackbox.config import Blackbox as CONFIG
from blackbox.utils.logger import log


class SanitizeReportMixin:
    """
    Mixin class helper to mask passwords and tokens set up in blackbox.yaml.

    Could mask any string with sensitive output because it has its own
    method to get all credentials from blackbox config.
    """

    def sanitize_output(self, sensitive_output: str) -> str:
        """Replace all self.config credentials values with ***** in any string."""
        # First of all let's combine list of secrets from config.
        sensitive_words = self._extract_secrets()

        # Now it's time to mask all secrets in our output.
        for sensitive_word in set(sensitive_words):
            sensitive_output = sensitive_output.replace(str(sensitive_word), "*****")

        # Emphasize output is clear now
        sanitized_output = sensitive_output
        log.success("Secrets are masked!")
        return sanitized_output

    @staticmethod
    def _extract_secrets() -> list:
        """Helper method to get secrets values."""
        sensitive_words = []
        # Blackbox_Type -> Kind -> Unique ID (str) -> Keys and Values (passwords and tokens)
        types = [CONFIG.databases.values(), CONFIG.storage.values(), CONFIG.notifiers.values(), ]

        for blackbox_type in types:
            for kind in blackbox_type:
                for unique_id in kind.values():
                    sensitive_words += list(unique_id.values())
        return sensitive_words



from blackbox.handlers.databases import MariaDB
from blackbox.handlers.databases._base import BlackboxDatabase


class MySQL(MariaDB, BlackboxDatabase):
    """A Database handler that will do a mysqldump for MySQL, backing up all tables."""

    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['blackbox',
 'blackbox.handlers',
 'blackbox.handlers.databases',
 'blackbox.handlers.notifiers',
 'blackbox.handlers.storage',
 'blackbox.utils']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.16.51,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'dropbox>=11.0.0,<12.0.0',
 'jinja2>=3.1.2,<4.0.0',
 'loguru>=0.5.3,<0.6.0',
 'pytelegrambotapi>=3.7.7,<4.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.25.1,<3.0.0',
 'single-source>=0.2.0,<0.3.0',
 'stone>=3.2.1,<4.0.0']

entry_points = \
{'console_scripts': ['blackbox = blackbox.cli:cli']}

setup_kwargs = {
    'name': 'blackbox-cli',
    'version': '2.3.1',
    'description': 'Tool for automatic backups of databases',
    'long_description': '![Lint & Test](https://github.com/lemonsaurus/blackbox/workflows/Lint%20&%20Test/badge.svg)\n![Docker Build](https://github.com/lemonsaurus/blackbox/workflows/Docker%20Build/badge.svg)\n[![PyPI version](https://badge.fury.io/py/blackbox-cli.svg)](https://pypi.org/project/blackbox-cli/)\n\n![blackbox](https://github.com/lemonsaurus/blackbox/raw/main/img/blackbox_banner.png)\n\nA simple service which magically backs up all your databases to all your\nfavorite cloud storage providers, and then notifies you.\n\nSimply create a config file, fill in some connection strings for your favorite\nservices, and schedule `blackbox` to run however often you want using something\nlike `cron`, or a Kubernetes CronJob.\n\n## Table of Contents\n\n- [Setup](#setup)\n- [Configuration](#configuration)\n- [Environment Variables](#environment-variables)\n- [Databases](#databases)\n    - [MongoDB](#mongodb)\n    - [PostgreSQL](#postgresql)\n    - [MariaDB](#mariadb)\n    - [MySQL](#mysql)\n    - [Redis](#redis)\n    - [Specify Storage Providers and Notifiers for each Database](#specify-storage-providers-and-notifiers-for-each-database)\n- [Storage Providers](#storage-providers)\n    - [S3](#s3)\n    - [Dropbox](#dropbox)\n- [Notifiers](#notifiers)\n    - [Discord](#discord)\n    - [Slack](#slack)\n    - [Telegram](#telegram)\n- [Rotation](#rotation)\n- [Cooldown](#cooldown)\n\n# Setup\n\nThis service can either be set up as a cron job (on UNIX systems), as a\nKubernetes CronJob, or scheduled in your favorite alternative scheduler.\n\n## Quick start\n\nRequires Python 3.9 or newer\n\n```sh\n# Install the CLI tool\npip install blackbox-cli\n\n# Create a configuration file\nblackbox --init\n\n# Run blackbox with a specific config file\nblackbox --config=/path/to/blackbox.yaml\n```\n\n### Setting up as a cron job\n\nAll you need to do to set it up as a cron job is clone this repo, create a\nconfig file (see below), and trigger `blackbox` to run automatically however\noften you want.\n\n```sh\ncrontab -e\n\n#run backup every hour\n0 */1 * * * blackbox --config path/to/blackbox.yml\n```\n\n### Setting it up as a Kubernetes CronJob\n\nTo set this up as a Kubernetes CronJob, you\'ll want three manifests and a\nsecret.\n\nBefore we start, you\'ll probably want to create a secret named `blackbox-secrets` where you expose\nenvironment variables containing stuff like passwords for your databases,\ncredentials for your storage, and webhooks as environment variables. We\'ll be\ninterpolating those into the config file.\n\nNext, we\'ll need a ConfigMap for the `blackbox.yaml` config file. See the\nConfiguration section below for more information on what to put inside this\nfile.\n\n```yaml\n# blackbox-configmap.yaml\napiVersion: v1\nkind: ConfigMap\nmetadata:\n  name: blackbox-config\ndata:\n  blackbox.yaml: |\n    databases:\n      mongodb:\n        main_mongodb:\n            connection_string: mongodb://{{ MONGO_INITDB_ROOT_USERNAME }}:{{ MONGO_INITDB_ROOT_PASSWORD }}@mongodb.default.svc.cluster.local:27017\n\n    storage:\n      s3:\n        main_s3:\n          bucket: blackbox\n          endpoint: my.s3.com\n          aws_access_key_id: {{ AWS_ACCESS_KEY_ID }}\n          aws_secret_access_key: {{ AWS_SECRET_ACCESS_KEY }}\n\n    notifiers:\n      discord:\n        main_discord:\n          webhook: {{ DISCORD_WEBHOOK }}\n\n    retention_days: 7\n```\n\nFinally, we need the CronJob itself. This one is configured to run once a day,\nat midnight.\n\n```yaml\n# cronjob.yaml\napiVersion: batch/v1\nkind: CronJob\nmetadata:\n  name: blackbox\nspec:\n  schedule: "@daily"\n  jobTemplate:\n    spec:\n      template:\n        spec:\n          containers:\n            - name: blackbox\n              image: lemonsaurus/blackbox\n              imagePullPolicy: Always\n              envFrom:\n                - secretRef:\n                    name: blackbox-secrets\n              # Tell blackbox where to find the config file.\n              env:\n              - name: BLACKBOX_CONFIG_PATH\n                value: "/blackbox/config_file/blackbox.yaml"\n              volumeMounts:\n                # Take care not to mount this in the root folder!\n                # That will replace everything in the root folder with\n                # the contents of this volume, which sucks.\n                - mountPath: /blackbox/config_file\n                  name: blackbox-config\n          volumes:\n            - name: blackbox-config\n              configMap:\n                name: blackbox-config\n          restartPolicy: OnFailure\n      backoffLimit: 3\n```\n\n# Configuration\n\n`blackbox` configuration is easy. You simply create a yaml\nfile, `blackbox.yaml`, which contains something like this:\n\nSee below for specific configuration information for each handler.\n\n```yaml\ndatabases:\n  postgres: # Database type \n    main_postgres: # Database identifier\n      # Configuration (see below for further information on specific databases)\n      username: username\n      password: password\n      host: host\n      port: port\n\n      # Optionally specify storage and notifiers to use\n      # You can specify them by type or identifier\n      # Use a string for a single specifier, a list for multiple specifiers\n      storage_providers:\n        - s3\n        - secondary_dropbox\n      notifiers: slack\n\n  redis:\n    main_redis:\n      password: password\n      host: host\n      port: port\n      # No specified storage and notifiers, so all storage and notifiers are used\n\nstorage:\n  s3: # Storage type\n    main_s3: # Storage identifier\n      bucket: bucket\n      endpoint: s3.endpoint.com\n    secondary_s3:\n      bucket: bucket\n      endpoint: s3.another_endpoint.com\n  dropbox:\n    main_dropbox:\n      access_token: XXXXXXXXXXX\n    secondary_dropbox:\n      access_token: XXXXXXXXXXX\n\nnotifiers:\n  discord: # Notifier type\n    main_discord: # Notifier identifier\n      webhook: https://discord.com/api/webhooks/797541821394714674/lzRM9DFggtfHZXGJTz3yE-MrYJ-4O-0AbdQg3uV2x4vFbu7HTHY2Njq8cx8oyMg0T3Wk\n  slack:\n    main_slack:\n      webhook: https://hooks.slack.com/services/XXXXXXXXXXX/XXXXXXXXXXX/XXXXXXXXXXXXXXXXXXX\n\nretention_days: 7\n```\n\nBlackbox will look for this file in the root folder by default, however you can\nprovide an alternative config file path by creating an environment variable\ncalled `BLACKBOX_CONFIG_PATH`, and set it to the absolute path of the file.\n\n```sh\nexport BLACKBOX_CONFIG_PATH=/var/my/favorite/fruit/blackbox.yaml\n```\n\nYou can also specify the location of this file when using the `blackbox` cli\ncommand.\n\n```sh\nblackbox --config=/path/to/blackbox.yaml\n```\n\n## Environment Variables\n\nThe `blackbox.yaml` will ✨ **magically interpolate** ✨ any environment\nvariables that exist in the environment where `blackbox` is running. This is\nvery useful if you want to keep your secrets in environment variables, instead\nof keeping them in the config file in plaintext.\n\n#### Example\n\nImagine your current config looks like this, but you want to move the username\nand password into environment variables.\n\n```yaml\ndatabases:\n  postgres:\n    main_postgres:\n      username: lemonsaurus\n      password: security-is-overrated\n      host: localhost\n      port: 5432\n```\n\nSo we\'ll create two environment variables like these:\n\n```sh\nexport POSTGRES_USERNAME=lemonsaurus\nexport POSTGRES_PASSWORD=security-is-overrated\n```\n\nAnd now we can make use of these environment variables by using double curly\nbrackets, like this:\n\n```yaml\ndatabases:\n  postgres:\n    main_postgres:\n      username: { { POSTGRES_USERNAME } }\n      password: { { POSTGRES_PASSWORD } }\n      host: localhost\n      port: 5432\n```\n\n## Databases\n\nRight now, this app supports **MongoDB**, **PostgreSQL 7.0 or higher**, **MariaDB**, **Redis** and **local storage archiving**. If\nyou need support for an additional database, consider opening a pull request to\nadd a new database handler.\n\nTo configure databases, add a section with this format:\n\n```yaml\ndatabases:\n  database_type:\n    # More than one of each database type can be configured\n    identifier_1:\n      field: value\n    identifier_2:\n      field: value\n  database_type:\n    ...\n```\n\nSee below for the specific database types available and fields required.\nIdentifiers can be any string of your choosing.\n\n### MongoDB\n\n- **Database Type**: `mongodb`\n- **Required fields**: `connection_string`\n- The `connection_string` field is in the\n  format `mongodb://username:password@host:port`\n- To restore from the backup,\n  use `mongorestore --gzip --archive=/path/to/backup.archive`\n\n```yaml\n  mongodb:\n    main_mongo:\n      connection_string: "mongodb://blackbox:blackbox@mongo:27017"\n```\n\n### PostgreSQL\n\n- **Database Type**: `postgres`\n- **Required fields**: `username`, `password`, `host`, `port`\n- To restore from the backup, use `psql -f /path/to/backup.sql`\n\n```yaml\n  postgres:\n    main_postgres:\n      username: blackbox\n      password: blackbox\n      host: postgres\n      port: "5432"\n```\n\n### MariaDB\n\n- **Database Type**: `mariadb`\n- **Required fields**: `username`, `password`, `host`, `port`\n- To restore from the backup, use `mysql -u <user> -p < db_backup.sql`\n\n```yaml\n  mariadb:\n    main_mariadb:\n      username: root\n      password: example\n      host: maria\n      port: "3306"\n```\n\n### MySQL\n\n- **Database Type**: `mysql`\n- **Required fields**: `username`, `password`, `host`, `port`\n- To restore from the backup, use `mysql -u <user> -p < db_backup.sql`\n\n```yaml\n  mysql:\n    main_mysql:\n      username: root\n      password: example\n      host: mysql\n      port: "3306"\n```\n\n### Redis\n\n- **Database Type**: `redis`\n- **Required fields**: `password`, `host`, `port`\n\n```yaml\n  redis:\n    main_redis:\n      password: blackbox\n      host: redis\n      port: "6379"\n```\n\n### Local storage\n\n- **Database type**: `localstorage`\n- **Required field**: `path`\n- **Optional field**: `compression_level`\n- The compression level must be an integer between 0 and 9.\n- The archive will contain the full structure, starting from the root folder.\n\n```yaml\n  localstorage:\n    main_localstorage:\n      path: /path/to/folder\n      compression_level: 7\n```\n\n#### To restore from the backup\n\n- Stop Redis server.\n- Turn off `appendonly` mode in Redis configuration (set to `no`).\n- Copy backup file to Redis working directory (`dir` in configuration) with\n  name that is defined in configuration key `dbfilename`.\n- Set backup permissions.\n\n```\nsudo chown redis:redis <path-to-redis-dump-file>\nsudo chmod 660 <path-to-redis-dump-file>\n```\n\n- Start Redis server.\n\nIf you want to re-enable `appendonly`:\n\n- Login with `redis-cli`.\n- Run `BGREWRITEAOF`.\n- Exit from Redis CLI (with `exit`).\n- Stop Redis server.\n- Set `appendonly` to `yes` in Redis configuration.\n- Start Redis server.\n\n### Specify Storage providers and Notifiers for each Database\n\nTo specify specific storage providers or notifiers for databases, add the\nfields `storage_providers` and `notifiers` under each database entry. The entry\ncan be a list or a string.\n\n```yaml\ndatabases:\n  postgres: # Database type \n    main_postgres: # Database identifier\n      username: username\n      password: password\n      host: host\n      port: port\n\n      storage_providers:\n        - s3\n        - secondary_dropbox\n      notifiers: slack\n```\n\nThe above example will backup `main_postgres` to every `s3` storage provider\nconfigured, as well as the storage provider with the\nidentifier `secondary_dropbox`. Then, only the `slack` notifier gets notified.\n\nThese fields are optional. If not given, all storage providers and all\nnotifiers will be used.\n\n## Storage providers\n\n**Blackbox** can work with different storage providers to save your logs and\nbackups - usually so that you can automatically store them in the cloud. Right\nnow we support **S3** and **Dropbox**.\n\nTo configure storage providers, add a section with this format:\n\n```yaml\nstorage:\n  storage_type:\n    # More than one of each storage provider type can be configured\n    identifier_1:\n      field: value\n    identifier_2:\n      field: value\n  storage_type:\n    ...\n```\n\n### S3\n\nWe support any S3 object storage bucket, whether it\'s from **AWS**, **Linode**\n, **DigitalOcean**, **Scaleway**, or another S3-compatible object storage\nprovider.\n\n**Blackbox** will respect the `retention_days` configuration setting and delete\nolder files from the S3 storage. Please note that if you have a bucket\nexpiration policy on your storage, **blackbox** will not do anything to disable\nit. So, for example, if your bucket expiration policy is 12 hours and blackbox\nis set to 7 `retention_days`, then your backups are all gonna be deleted after\n12 hours unless you disable your policy.\n\n#### S3 configuration\n\n- **Storage Type**: `s3`\n- **Required fields**: `bucket`, `endpoint`\n- **Optional fields**: `aws_access_key_id`, `aws_secret_access_key`\n- The `endpoint` field can look something like\n  this: `s3.eu-west-1.amazonaws.com`\n\n#### Credentials\n\nTo upload stuff to S3, you\'ll need credentials. Your **AWS credentials** can be\nprovided in several ways. This is the order in which blackbox looks for them:\n\n- First, we look for the optional fields in the s3 configuration,\n  called `aws_access_key_id` and `aws_secret_access_key`.\n- If these are not found, we\'ll check if the `AWS_ACCESS_KEY_ID`\n  and `AWS_SECRET_ACCESS_KEY` environment variables are declared in the local\n  environment where Blackbox is running.\n- If we can\'t find these, we\'ll look for an `.aws/config` file in the local\n  environment.\n- NOTE: If the bucket is public, no credentials are necessary.\n\n### Dropbox\n\n- **Storage Type**: `dropbox`\n- **Required fields**: `access_token`\n- **Optional fields**: `upload_directory`\n\nThe Dropbox storage handler needs a user access token in order to work. To get\none, do the following:\n\n- Create a Dropbox account (if you don\'t already have one).\n- Go to https://dropbox.com/developers\n- Create a new application with App Folder access. **Do not give it full\n  access**, as this may have dangerous, destructive consequences if configured\n  incorrectly.\n\nYou can also define a custom location (root is App Folder) using the\n`upload_directory` optional parameter. This **should** begin with slash and \n**must** end with slash. Default is root.\n\n## Notifiers\n\n`blackbox` also implements different _notifiers_, which is how it reports the\nresult of one of its jobs to you. Right now we only support **Discord** and **Slack**\nand **Telegram**, but if you need a specific notifier, feel free to open an issue.\n\nTo configure notifiers, add a section with this format:\n\n```yaml\nnotifiers:\n  notifier_type:\n    # More than one of each notifier type can be configured\n    identifier_1:\n      field: value\n    identifier_2:\n      field: value\n  notifier_type:\n    ...\n```\n\n### Discord\n\n- **Notifier Type**: `discord`\n- **Required fields**: `webhook`\n- The `webhook` field usually looks\n  like `https://discord.com/api/webhooks/797541821394714674/lzRM9DFggtfHZXGJTz3yE-MrYJ-4O-0AbdQg3uV2x4vFbu7HTHY2Njq8cx8oyMg0T3Wk`\n- We also support `ptb.discord.com` and `canary.discord.com` webhooks.\n\n![blackbox](https://github.com/lemonsaurus/blackbox/raw/main/img/blackbox_discord.png)\n![blackbox](https://github.com/lemonsaurus/blackbox/raw/main/img/blackbox_discord_2.png)\n\n### Slack\n\n- **Notifier Type**: `slack`\n- **Required fields**: `webhook`\n- The `webhook` field usually looks\n  like `https://hooks.slack.com/services/XXXXXXXXXXX/XXXXXXXXXXX/XXXXXXXXXXXXXXXXXXX`\n\nSlack notifiers have 2 styles: legacy attachment (default) and modern Block Kit\nversion. To enable Block Kit version, set the optional field `use_block_kit` to\nanything.\n\nDefault:\n\n![blackbox](https://github.com/lemonsaurus/blackbox/raw/main/img/blackbox_slack_default_success.png)\n![blackbox](https://github.com/lemonsaurus/blackbox/raw/main/img/blackbox_slack_default_fail.png)\n\nModern:\n\n![blackbox](https://github.com/lemonsaurus/blackbox/raw/main/img/blackbox_slack_modern_success.png)\n![blackbox](https://github.com/lemonsaurus/blackbox/raw/main/img/blackbox_slack_modern_fail.png)\n\n\n### Telegram\n\n- **Notifier Type**: `telegram`\n- **Required fields**: `token`, `chat_id`\n- YAML will look like this:\n```notifiers:\n  telegram:\n    telegram_1:\n      token: {{ TELEGRAM_TOKEN }}\n      chat_id: {{ TELEGRAM_CHAT_ID }}\n```\n- You can create a bot and get a bot token using the `BotFather` account in Telegram. Follow [these instructions](https://core.telegram.org/bots#6-botfather).\n- You can find your `chat_id` by using the `userinfobot` account in Telegram. Just `/start` the bot.\n- Do not forget to `/start` your own bot to grant sending permissions.\n\n![blackbox](https://github.com/lemonsaurus/blackbox/raw/main/img/blackbox_telegram_success.png)\n![blackbox](https://github.com/lemonsaurus/blackbox/raw/main/img/blackbox_telegram_fail.png)\n\n\n## Rotation\n\nBy default, `blackbox` will automatically remove all backup files older than 7\ndays in the folder you configure for your storage provider. To determine if\nsomething is a backup file or not, it will use a regex pattern that corresponds\nwith the default file it saves, for\nexample `blackbox-postgres-backup-11-12-2020.sql`.\n\nYou can configure the number of days before rotating by altering\nthe `retention_days` parameter in `blackbox.yaml`.\n\n\n## Cooldown\n\nBy default, `blackbox` will send all notification at every backup attempt. \nYou can specify a `cooldown` period in `blackbox.yaml` during which all notifications will be muted.\nThis option will not mute failed backups.\n\n### Example usage\n\n```yaml\ncooldown: 120s\n```\n```yaml\ncooldown: 3 hours\n```\n```yaml\ncooldown: 2 days 4 hours\n```\n```yaml\ncooldown: 4h 32M 16s\n```\n',
    'author': 'Leon Sandøy',
    'author_email': 'leon.sandoy@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/lemonsaurus/blackbox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from abc import ABC

from blackbox.exceptions import MissingFields
from blackbox.utils.mixins import SanitizeReportMixin


class BlackboxHandler(ABC, SanitizeReportMixin):
    """An abstract base handler."""

    required_fields = ()
    handler_type = ""

    def __init__(self, **config):
        self.config = config
        self._validate_config()

    def _validate_config(self):
        """Ensure required configuration fields were passed to the handler."""
        handler_name = self.__class__.__name__
        missing_fields = [field for field in self.required_fields if field not in self.config]

        if missing_fields:
            raise MissingFields(self.handler_type, handler_name, self.config.get("id"),
                                missing_fields)

    def teardown(self) -> None:
        """
        This gets called at the end of the handlers life.

        It doesn't need to be implemented, but it can be if you need to do some form of
        teardown or cleanup after your handler has done whatever it is designed to do.
        """
        return None



import json
import os
import re
from datetime import datetime
from pathlib import Path
from typing import Optional

from dateutil.relativedelta import relativedelta

from blackbox.utils.logger import log


_DURATION_REGEX = re.compile(
    r"((?P<years>\d+?) ?(years|year|Y|y) ?)?"
    r"((?P<months>\d+?) ?(months|month|m) ?)?"
    r"((?P<weeks>\d+?) ?(weeks|week|W|w) ?)?"
    r"((?P<days>\d+?) ?(days|day|D|d) ?)?"
    r"((?P<hours>\d+?) ?(hours|hour|H|h) ?)?"
    r"((?P<minutes>\d+?) ?(minutes|minute|M) ?)?"
    r"((?P<seconds>\d+?) ?(seconds|second|S|s))?"
)


def parse_config_cooldown(cooldown) -> Optional[relativedelta]:
    """Convert human period to relativedata."""
    match = _DURATION_REGEX.fullmatch(cooldown)
    if not match:
        return None

    duration_dict = {unit: int(amount) for unit, amount in match.groupdict(default=0).items()}
    delta = relativedelta(**duration_dict)

    return delta


def should_notify(last_send: datetime, delta: relativedelta) -> bool:
    """Check if we can send now."""
    return True if datetime.now() - delta >= last_send else False


def write_config():
    """Write down successful notification."""
    data = {'last_send': datetime.now().strftime("%m/%d/%Y, %H:%M:%S")}
    log.debug(f"Sending notification at {data}")
    with open(get_project_root() / 'notify.json', 'w', encoding='utf-8') as f:
        json.dump(data, f, ensure_ascii=False, indent=4)


def read_config() -> dict:
    """Read last notification time."""
    log.debug('Found json, reading it...')
    with open(get_project_root() / 'notify.json') as infile:
        data = json.load(infile)
    return data


def get_project_root() -> Path:
    """Fancy way to retrieve project root."""
    return Path(__file__).parent.parent


def is_on_cooldown(cooldown) -> bool:
    """Check if we can send notification, main function."""
    log.debug(f'Cooldown period set to {cooldown}')
    delta = parse_config_cooldown(cooldown)
    if os.path.exists(get_project_root() / 'notify.json'):
        data = read_config()
        last_send = datetime.strptime(data['last_send'], "%m/%d/%Y, %H:%M:%S")
        if should_notify(last_send, delta):
            write_config()
            return False
        log.debug('On cooldown')
        return True
    else:
        write_config()
        return False



from ._base import BlackboxDatabase
from .localstorage import LocalStorage
from .mariadb import MariaDB
from .mongodb import MongoDB
from .mysql import MySQL
from .postgres import Postgres
from .redis import Redis
